# -*- coding: UTF-8 -*-
import dataclasses
import io
import typing
from collections import UserString

import orjson
from starlette.background import BackgroundTask
from starlette.responses import Response, StreamingResponse, ContentStream

type BaseTypeAnnotation = str | int | float | bool | None


@dataclasses.dataclass(slots=True, kw_only=True)
class _CustomDataType:
    type_class: type
    encoder: typing.Callable[[typing.Any], BaseTypeAnnotation]


_CUSTOM_DATA_TYPE_LIST = [_CustomDataType(type_class=UserString, encoder=str)]


def register_custom_data_types(
        *, type_class: type,
        encoder: typing.Callable[[typing.Any], BaseTypeAnnotation],
) -> None:
    for custom_data_type in _CUSTOM_DATA_TYPE_LIST:
        if custom_data_type.type_class == type_class:
            raise ValueError(f'Type class {type_class} already registered')
    _CUSTOM_DATA_TYPE_LIST.append(
        _CustomDataType(type_class=type_class, encoder=encoder))


def orjson_dumps_default(obj):
    for custom_data_type in _CUSTOM_DATA_TYPE_LIST:
        if isinstance(obj, custom_data_type.type_class):
            return custom_data_type.encoder(obj)
    raise TypeError(f'Object of type {type(obj)} is not JSON serializable')


class ORJSONResponse(Response):
    media_type = 'application/json'

    def render(self, content: typing.Any) -> bytes:
        return orjson.dumps(
            content, option=orjson.OPT_STRICT_INTEGER,
            default=orjson_dumps_default)


class StreamResponse(StreamingResponse):
    pass


async def _make_bytes_generator(
        *, content: bytes, chunk_size: int,
) -> typing.AsyncGenerator[bytes, None]:
    length = len(content)
    seek = 0
    while seek < length:
        yield content[seek:seek + chunk_size]
        seek += chunk_size


async def _make_bio_generator(
        *, content: io.BytesIO, chunk_size: int,
) -> typing.AsyncGenerator[bytes, None]:
    content.seek(0)
    while True:
        chunk = content.read(chunk_size)
        if not chunk:
            break
        yield chunk
    content.close()


class FileBytesResponse(StreamResponse):

    def __init__(
            self,
            content: ContentStream | bytes,
            status_code: int = 200,
            headers: typing.Mapping[str, str] | None = None,
            media_type: str | None = None,
            background: BackgroundTask | None = None,
            chunk_size: int = 1024 * 1024,
    ) -> None:
        super().__init__(
            content=content, status_code=status_code, headers=headers,
            media_type=media_type, background=background)
        if isinstance(content, bytes):
            self.headers['Content-Length'] = str(len(content))
            self.body_iterator = _make_bytes_generator(
                content=content, chunk_size=chunk_size)
        if isinstance(content, io.BytesIO):
            content.seek(0, io.SEEK_END)
            self.headers['Content-Length'] = str(content.tell())
            self.body_iterator = _make_bio_generator(
                content=content, chunk_size=chunk_size)



# -*- coding: UTF-8 -*-

import abc
import dataclasses
import functools
import inspect
import typing

import orjson
from psycopg import AsyncConnection, AsyncCursor
from psycopg.rows import Row
from psycopg.sql import SQL
from psycopg.types.json import set_json_dumps, set_json_loads
from psycopg_pool import AsyncConnectionPool

set_json_loads(orjson.loads)
set_json_dumps(orjson.dumps)

_NAMED_POOL_DICT: dict[str, AsyncConnectionPool] = dict()

type ConfigureFunc = typing.Callable[
    [AsyncConnection], typing.Awaitable | None]


async def init(
        *, name: str, dsn: str, pool_size: int = 16, open_timeout: int = 15,
        configure_funcs: typing.Sequence[ConfigureFunc] = None,
) -> None:
    if name in _NAMED_POOL_DICT:
        raise ValueError(f'Pool "{name}" already exists')
    if pool_size < 1:
        raise ValueError('Pool size should be greater than 0')
    minsize, maxsize = pool_size, pool_size
    if minsize > 4:
        minsize = 4
    kwargs = dict(autocommit=False)
    init_kwargs = dict()
    if configure_funcs:
        async def _configure(conn: AsyncConnection):
            for func in configure_funcs:
                if inspect.iscoroutinefunction(func):
                    await func(conn)
                else:
                    func(conn)

        init_kwargs['configure'] = _configure
    pool = AsyncConnectionPool(
        conninfo=dsn, min_size=minsize, max_size=maxsize, open=False,
        name=name, kwargs=kwargs, **init_kwargs)
    await pool.open(wait=True, timeout=open_timeout)
    _NAMED_POOL_DICT[name] = pool


async def close_all():
    for _, pool in _NAMED_POOL_DICT.items():
        await pool.close()
    _NAMED_POOL_DICT.clear()


def with_postgres(*, name: str, transaction: bool = False):
    def wrapper(func):
        argspec = inspect.getfullargspec(func)
        if all(map(lambda x: 'cursor' not in x,
                   (argspec.args, argspec.kwonlyargs))):
            raise SyntaxError('`cursor` is a required argument')

        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            if name not in _NAMED_POOL_DICT:
                raise SyntaxError(f'Pool "{name}" not found')
            if 'cursor' in kwargs:
                raise SyntaxError('`cursor` is a reserved argument')
            async with _NAMED_POOL_DICT[name].connection() as conn:
                async with conn.cursor() as cursor:
                    kwargs['cursor'] = cursor
                    if transaction:
                        async with conn.transaction():
                            result = await func(*args, **kwargs)
                    else:
                        result = await func(*args, **kwargs)
            return result

        return wrapped

    return wrapper


class MusoSQLHelper:

    def __init__(self, base_model: type | None = None):
        try:
            assert dataclasses.is_dataclass(base_model) or base_model is None
        except AssertionError:
            raise SyntaxError(
                f'Base model "{base_model}" should be a dataclass or None')
        self._base_model = base_model

    def _gen_fields(self, dataclass: type) -> tuple[list[str], list[str]]:
        if self._base_model and issubclass(dataclass, self._base_model):
            default_fields = [
                i.strip('_')
                for i in inspect.get_annotations(self._base_model).keys()]
        else:
            default_fields = list()
        custom_fields = [
            i.strip('_') for i in inspect.get_annotations(dataclass).keys()]
        return default_fields, custom_fields

    def gen_select_base(self, *, dataclass: type, table_name: str) -> str:
        default_fields, custom_fields = self._gen_fields(dataclass)
        return (f'SELECT {", ".join(default_fields + custom_fields)} '
                f'FROM {table_name}')

    def gen_get(self, *, dataclass: type, table_name: str) -> SQL:
        default_fields, custom_fields = self._gen_fields(dataclass)
        return self.make_sql(sql_str=(
            f'SELECT {", ".join(default_fields + custom_fields)} '
            f'FROM {table_name} WHERE id=%s'))

    def gen_insert(self, *, dataclass: type, table_name: str) -> SQL:
        _, custom_fields = self._gen_fields(dataclass)
        return self.make_sql(sql_str=(
            f'INSERT INTO {table_name} ({", ".join(custom_fields)}) '
            f'VALUES ({", ".join(["%s"] * len(custom_fields))})'))

    def gen_insert_returning(
            self, *, dataclass: type, table_name: str) -> SQL:
        default_fields, custom_fields = self._gen_fields(dataclass)
        return self.make_sql(sql_str=(
            f'INSERT INTO {table_name} ({", ".join(custom_fields)}) '
            f'VALUES ({", ".join(["%s"] * len(custom_fields))}) '
            f'RETURNING {", ".join(default_fields + custom_fields)}'))

    def gen_remove(self, *, table_name: str) -> SQL:
        return self.make_sql(sql_str=(
            f'UPDATE {table_name} SET removed=%s, updated_at=NOW() '
            f'WHERE id=%s AND removed=%s'))

    @staticmethod
    def make_sql(*, sql_str: str) -> SQL:
        return SQL(sql_str.strip())     # noqa


@dataclasses.dataclass(slots=True, kw_only=True)
class BaseInsertModel(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def as_sql_params(self) -> tuple:
        ...


type DefaultCursorAnnotation = AsyncCursor[AsyncConnection['Row'], 'Row']


class BaseDataFetcher[_T](metaclass=abc.ABCMeta):
    _MODEL_CLASS: type[_T]

    _TABLE_NAME: str
    _SELECT_BASE: str

    _sql_get: SQL

    @classmethod
    async def get(
            cls, *, cursor: DefaultCursorAnnotation, id_: int,
    ) -> _T | None:
        await cursor.execute(query=cls._sql_get, params=(id_,))
        result: Row | None = await cursor.fetchone()
        return cls._MODEL_CLASS(*result) if result else None

    _sql_remove: SQL

    @classmethod
    async def remove(
            cls, *, cursor: DefaultCursorAnnotation, id_: int,
    ) -> None:
        await cursor.execute(query=cls._sql_remove, params=(True, id_, False))



# -*- coding: UTF-8 -*-

import orjson
import tzlocal
from marshmallow import Schema, fields


class BaseSchema(Schema):
    class Meta:
        render_module = orjson


class LimitOffsetSchema(BaseSchema):
    limit = fields.Integer(required=True, description='限制数量')
    offset = fields.Integer(required=True, description='偏移量')


class PaginatorSchema(BaseSchema):
    page_no = fields.Integer(
        data_key='pageNo', required=True, description='页码')
    page_size = fields.Integer(
        data_key='pageSize', required=True, description='每页数量')


class EmptyResponseSchema(BaseSchema):
    code = fields.Integer(missing=0, allow_none=False, dump_default=0)
    msg = fields.String(missing='', dump_default='')
    data = fields.Raw(missing=None, allow_none=True)


class LocalDateTimeField(fields.DateTime):

    def __init__(self, ignore_microsecond: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ignore_microsecond = ignore_microsecond

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return None
        if self.ignore_microsecond:
            value = value.replace(microsecond=0)
        return super()._serialize(value.astimezone(
            tz=tzlocal.get_localzone()), attr, obj, **kwargs)


class StarletteFileField(fields.Field):

    def __init__(self, content_type_list: list[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content_type_list = content_type_list


class EnumNameField(fields.Field):

    def __init__(self, enum_class, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enum_class = enum_class

    def _serialize(self, value, attr, obj, **kwargs):
        return value.name

    def _deserialize(self, value, attr, data, **kwargs):
        return self.enum_class[value]


class EnumValueField(fields.Field):

    def __init__(self, enum_class, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enum_class = enum_class

    def _serialize(self, value, attr, obj, **kwargs):
        return value.value

    def _deserialize(self, value, attr, data, **kwargs):
        return self.enum_class(value)



# -*- coding: UTF-8 -*-
import functools
import inspect

from redis.asyncio import BlockingConnectionPool, StrictRedis

_NAMED_POOL_DICT: dict[str, BlockingConnectionPool] = dict()


async def init(
        *, name: str, url: str, max_connections: int = 16, timeout: int = 10,
) -> None:
    if name in _NAMED_POOL_DICT:
        raise ValueError(f'Pool "{name}" already exists')
    if max_connections < 1:
        raise ValueError('Pool size should be greater than 0')
    pool = BlockingConnectionPool.from_url(
        url=url, max_connections=max_connections, timeout=timeout, protocol=3)
    connection = await pool.get_connection('')
    await pool.release(connection)
    _NAMED_POOL_DICT[name] = pool


async def close_all():
    for _, pool in _NAMED_POOL_DICT.items():
        await pool.aclose()
    _NAMED_POOL_DICT.clear()


type RedisAnnotation = StrictRedis


def with_redis(*, name: str):
    def decorator(func):
        argspec = inspect.getfullargspec(func)
        if all(map(lambda x: 'redis' not in x,
                   (argspec.args, argspec.kwonlyargs))):
            raise SyntaxError('`redis` is a required argument')

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            pool = _NAMED_POOL_DICT.get(name)
            if not pool:
                raise SyntaxError(f'Pool {name} not found')
            if 'redis' in kwargs:
                raise SyntaxError('`redis` is a reserved argument')
            async with StrictRedis(connection_pool=pool) as redis:
                kwargs['redis'] = redis
                result = await func(*args, **kwargs)
            return result

        return wrapper

    return decorator
